from typing import Any, Dict, Tuple

import dask.array as da
import h5py
from napari.layers.utils.layer_utils import calc_data_range
import numpy as np
import pandas as pd

from ...contour_manager import ContourManager
from ...images.masks import make_scalar_mask


def create_cell_mask(
    stat: Dict[str, Any], n_rows: int, n_cols: int, allow_overlap: bool = False
) -> Tuple[np.ndarray, np.ndarray]:
    """
    from suite2p:
    https://github.com/MouseLand/suite2p/blob/d08f1a82561330d45142c0b6192ec779254e4533/suite2p/detection/masks.py#L22

    creates cell masks for ROIs in stat and computes radii

    Parameters
    ----------
    stat : dictionary 'ypix', 'xpix', 'lam'
    n_rows : y size of frame
    n_cols : x size of frame
    allow_overlap : whether or not to include overlapping pixels in cell masks

    Returns
    -------
    cell_masks : len ncells, each has tuple of pixels belonging to each cell and weights
    lam_normed
    """
    mask = ... if allow_overlap else ~stat["overlap"]
    cell_mask = np.ravel_multi_index(
        (stat["ypix"], stat["xpix"]), (n_rows, n_cols)
    )
    cell_mask = cell_mask[mask]
    lam = stat["lam"][mask]
    lam_normed = lam / lam.sum() if lam.size > 0 else np.empty(0)
    return cell_mask, lam_normed


def create_cell_labels_ims(
    stat_all: list, is_cell: np.ndarray, ops: dict,
) -> Tuple[np.ndarray, np.ndarray]:
    """Create a label images for the good and bad cells

    Parameters
    ----------
    stat_all : list
        The list of stat for each cell loaded from suite2p
    is_cell : np.ndarray
        This is the contents of the iscell.npy file
    ops : dict
        The options and intermediate outputs from s2p

    Returns
    -------
    good_cell_mask : np.ndarray
        A label image with each cell that passed QC in iscell.npy. Has the same shape
        as the source image.
    bad_cell_mask : np.ndarray
        A label image with each cell that failed QC in iscell.npy. Has the same shape
        as the source image.
    """

    # image shape
    n_rows = ops["Ly"]
    n_cols = ops["Lx"]

    cell_masks = [
        create_cell_mask(
            stat,
            n_rows=n_rows,
            n_cols=n_cols,
            allow_overlap=ops["allow_overlap"],
        )[0]
        for stat in stat_all
    ]

    good_cell_mask = np.zeros((n_rows, n_cols))
    bad_cell_mask = np.zeros((n_rows, n_cols))
    for cell_index, cell in enumerate(is_cell):
        cell_mask_indices = np.unravel_index(
            cell_masks[cell_index], (ops["Ly"], ops["Lx"])
        )
        if cell[0] == 1:
            good_cell_mask[cell_mask_indices] = cell_index
        else:
            bad_cell_mask[cell_mask_indices] = cell_index

    return good_cell_mask, bad_cell_mask


def create_cell_mask_indices(stat_all: list, ops: dict):
    n_rows = ops["Ly"]
    n_cols = ops["Lx"]

    cell_masks = [
        create_cell_mask(
            stat,
            n_rows=n_rows,
            n_cols=n_cols,
            allow_overlap=ops["allow_overlap"],
        )[0]
        for stat in stat_all
    ]

    cell_mask_indices = [
        np.vstack(np.unravel_index(cell, (ops["Ly"], ops["Lx"]))).T
        for cell in cell_masks
    ]

    return cell_mask_indices


def s2p_reader(
    pipeline_params, image_path, snr_path, trace_path, cell_path, spikes_path
):
    stat_all = np.load(pipeline_params, allow_pickle=True)
    is_cell = np.load(cell_path, allow_pickle=True)
    f_traces = np.load(trace_path, allow_pickle=True)
    spikes = np.load(spikes_path)

    # load the shifts
    ops = np.load("ops.npy", allow_pickle=True).item()
    y_offset = ops["yoff"]
    x_offset = ops["xoff"]
    offsets = np.vstack((y_offset, x_offset)).T

    def translate_slice(array, offsets, block_info=None):
        if block_info is not None:
            array_location = block_info[None]["array-location"]
            t_slice = array_location[0][0]
            registered_array = np.roll(
                array,
                (
                    -np.int16(offsets[t_slice][0]),
                    -np.int16(offsets[t_slice][1]),
                ),
                axis=(1, 2),
            )

        else:
            registered_array = array

        return registered_array

    f = h5py.File(image_path, "r")
    im = f["MSession_0/MUnit_0/Channel_0"]
    im_shape = im.shape
    da_im = da.from_array(im, chunks=(1, im_shape[-2], im_shape[-1]))
    data_range = calc_data_range(da_im)
    im_registered = da_im.map_blocks(translate_slice, offsets=offsets)

    # load SNR
    snr_df = pd.read_csv(snr_path)
    snr = snr_df["0"].values

    cell_mask_indicies = create_cell_mask_indices(stat_all, ops)
    initial_state = is_cell[:, 0].astype(np.bool)
    contour_manager = ContourManager(
        cell_mask_indicies,
        initial_state=initial_state,
        im_shape=(im_shape[-2], im_shape[-1]),
    )

    snr_mask = make_scalar_mask(
        cell_mask_indicies, im_shape=(im_shape[-2], im_shape[-1]), values=snr
    )

    return (
        im_registered,
        data_range,
        contour_manager,
        f_traces,
        snr,
        snr_mask,
        spikes,
        is_cell,
    )



from typing import Optional

from qtpy.QtWidgets import QWidget, QVBoxLayout
import pyqtgraph as pg
import numpy as np


class LinePlot(QWidget):
    def __init__(
        self,
        x: np.ndarray,
        y: np.ndarray,
        xlabel: Optional[str] = None,
        ylabel: Optional[str] = None,
        events: np.ndarray = None,
        parent=None,
    ):
        super(LinePlot, self).__init__(parent)
        self.vbox = QVBoxLayout()
        self._curves = []

        self.add_plot(x, y, xlabel=xlabel, ylabel=ylabel)
        if events is not None:
            self.add_events(events, y)
        else:
            self._events_plot = None
        self.setLayout(self.vbox)

    def add_plot(
        self,
        x,
        y,
        xlabel: Optional[str] = None,
        ylabel: Optional[str] = None,
        vert_line: bool = True,
    ):

        new_plot = pg.plot()
        vline = pg.InfiniteLine(angle=90, movable=False)
        new_plot.addItem(vline)

        # add axis labels
        if xlabel is not None:
            new_plot.setLabel("bottom", xlabel)
        if ylabel is not None:
            new_plot.setLabel("left", ylabel)

        self._plot = new_plot
        # self._curves.append(self._plot.plot(data))
        curve_item = pg.PlotCurveItem(x, y)
        self._plot.addItem(curve_item)

        self._curves.append(curve_item)

        self._vert_line = vline

        self.vbox.addWidget(new_plot)

    def add_events(self, events, f_trace):
        x = np.argwhere(events)
        y = f_trace[x]
        pos = np.hstack((x, y))
        if len(pos) > 0:
            self._events_plot = pg.ScatterPlotItem(pos=pos)
        else:
            self._events_plot = pg.ScatterPlotItem()
        self._plot.addItem(self._events_plot)

    def update_vline(self, new_pos):
        self._vert_line.setValue(new_pos)

    def plot(self, data, events=None):
        self.clear()
        self._plot.plot(data)

        if events is not None:
            if len(events) > 0 and self._events_plot is not None:
                self.add_events(events, data)

    def clear(self):
        self._plot.clear()
        vline = pg.InfiniteLine(angle=90, movable=False)
        self._plot.addItem(vline)
        self._vert_line = vline


class Histogram(LinePlot):
    def __init__(
        self,
        x: np.ndarray,
        y: np.ndarray,
        xlabel: Optional[str] = None,
        ylabel: Optional[str] = None,
        parent=None,
    ):
        super().__init__(
            x, y, xlabel=xlabel, ylabel=ylabel, events=None, parent=parent
        )

    def add_plot(
        self,
        x,
        y,
        xlabel: Optional[str] = None,
        ylabel: Optional[str] = None,
        vert_line: bool = True,
    ):
        new_plot = pg.plot()
        vline = pg.InfiniteLine(angle=90, movable=True)
        new_plot.addItem(vline)

        # add axis labels
        if xlabel is not None:
            new_plot.setLabel("bottom", xlabel)
        if ylabel is not None:
            new_plot.setLabel("left", ylabel)

        self._plot = new_plot
        # self._curves.append(self._plot.plot(data))
        curve_item = pg.PlotCurveItem(
            x, y, stepMode=True, fillLevel=0, brush=(0, 0, 255, 150)
        )
        self._plot.addItem(curve_item)

        self._curves.append(curve_item)

        self._vert_line = vline

        self.vbox.addWidget(new_plot)

    def connect_line_dragged(self, func):
        self._vert_line.sigPositionChanged.connect(func)



import numpy as np


def calc_data_range(data):
    """Calculate range of data values. If all values are equal return [0, 1].
    Parameters

    Adapted from
    https://github.com/napari/napari/blob/master/napari/layers/utils/layer_utils.py
    ----------
    data : array
        Data to calculate range of values over.
    Returns
    -------
    values : list of float
        Range of values.
    """
    bottom_plane_idx = (0,) * (data.ndim - 2)
    middle_plane_idx = tuple(s // 2 for s in data.shape[:-2])
    top_plane_idx = tuple(s - 1 for s in data.shape[:-2])
    idxs = [bottom_plane_idx, middle_plane_idx, top_plane_idx]
    reduced_data = [
        [np.max(data[idx]) for idx in idxs],
        [np.min(data[idx]) for idx in idxs],
    ]

    min_val = np.min(reduced_data)
    max_val = np.max(reduced_data)

    if min_val == max_val:
        min_val = 0
        max_val = 1
    return [float(min_val), float(max_val)]



from typing import Union

import numpy as np


class ContourManager:
    def __init__(
        self,
        contours: list,
        im_shape: tuple,
        initial_state: Union[str, np.ndarray] = "good",
    ):
        self._contours = contours
        self._contour_labels = np.arange(1, len(contours) + 1)
        self._im_shape = im_shape

        if isinstance(initial_state, str):
            if initial_state == "good":
                self._good_contour = np.ones((len(contours),), dtype=np.bool)
            else:
                self._good_contour = np.zeros((len(contours),), dtype=np.bool)
        elif isinstance(initial_state, np.ndarray):
            self._good_contour = initial_state
        else:
            raise TypeError("initial_state should be a string or numpy array")

        self._selected_contours = {}

    @property
    def contours(self) -> list:
        return self._contours

    @contours.setter
    def contours(self, contours: list):
        self._contours = contours

    @property
    def good_contour(self) -> np.ndarray:
        return self._good_contour

    @good_contour.setter
    def good_contour(self, good_contour):
        self._good_contour = good_contour

    @property
    def accepted_contours(self) -> list:
        good_contour_indices = np.argwhere(self.good_contour)
        return [self.contours[i[0]] for i in good_contour_indices]

    @property
    def rejected_contours(self) -> list:
        bad_contour_indices = np.argwhere(np.logical_not(self.good_contour))
        return [self.contours[i[0]] for i in bad_contour_indices]

    @property
    def selected_contours(self) -> set:
        return self._selected_contours

    @selected_contours.setter
    def selected_contours(
        self, selected_contours: Union[list, np.ndarray, set]
    ):
        self._selected_contours = set(selected_contours)

    def make_accepted_mask(self):
        accepted_contours_image = np.zeros(self._im_shape, dtype=np.uint16)
        labels = self._contour_labels[self.good_contour]
        for label, cont in zip(labels, self.accepted_contours):
            accepted_contours_image[
                np.round(cont[:, 0]).astype("int"),
                np.round(cont[:, 1]).astype("int"),
            ] = label

        return accepted_contours_image

    def make_rejected_mask(self):
        rejected_contours_image = np.zeros(self._im_shape, dtype=np.uint16)
        labels = self._contour_labels[np.logical_not(self.good_contour)]
        for label, cont in zip(labels, self.rejected_contours):
            rejected_contours_image[
                np.round(cont[:, 0]).astype("int"),
                np.round(cont[:, 1]).astype("int"),
            ] = label

        return rejected_contours_image



from typing import Optional

import numpy as np
from qtpy.QtWidgets import (
    QWidget,
    QGridLayout,
    QLabel,
    QLineEdit,
    QHBoxLayout,
    QSpacerItem,
)

from .plots import Histogram


class HistogramWidget(QWidget):
    def __init__(
        self,
        x: np.ndarray,
        y: np.ndarray,
        xlabel: Optional[str] = None,
        ylabel: Optional[str] = None,
        parent=None,
    ):
        super(HistogramWidget, self).__init__(parent)

        # set the width of the histogram plot to match the napari layer control width
        self.setMinimumWidth(240)
        self.setMaximumWidth(240)

        self.hist_plot = Histogram(
            x, y, xlabel=xlabel, ylabel=ylabel, parent=self
        )
        self.hist_plot.setMaximumWidth(230)

        self.thresh_text = QLineEdit()
        self.text_layout = QHBoxLayout()
        self.text_layout.addWidget(QLabel("SNR threshold:"))
        self.text_layout.addWidget(self.thresh_text)
        self.text_layout.addItem(QSpacerItem(5, 1))

        self.grid_layout = QGridLayout(self)
        self.grid_layout.setContentsMargins(0, 0, 0, 0)
        self.grid_layout.setSpacing(2)
        # self.grid_layout.setColumnMinimumWidth(0, 86)
        # self.grid_layout.setColumnStretch(1, 1)
        self.grid_layout.addWidget(self.hist_plot, 0, 0, 4, 6)
        self.grid_layout.addLayout(self.text_layout, 4, 0)
        self.grid_layout.setRowStretch(5, 1)
        self.grid_layout.setColumnStretch(1, 1)
        self.setLayout(self.grid_layout)

        self._on_hist_thresh_change()

        # connect events
        self.hist_plot.connect_line_dragged(self._on_hist_thresh_change)
        self.thresh_text.returnPressed.connect(self._on_thresh_text_change)

    def _on_hist_thresh_change(self):
        hist_thresh_value = self.hist_plot._vert_line.getPos()[0]
        self.thresh_text.setText(f"{hist_thresh_value:.2f}")

    def _on_thresh_text_change(self):
        hist_thresh_value = float(self.thresh_text.text())
        self.hist_plot._vert_line.setValue(hist_thresh_value)



from typing import Optional

import napari
import numpy as np

from .contour_manager import ContourManager
from .qt.plots import LinePlot
from .qt.dock_widgets import HistogramWidget

from .images.masks import make_scalar_mask


def calcium_curator(
    img: np.ndarray,
    data_range,
    contour_manager: ContourManager,
    f: Optional[np.ndarray] = None,
    spikes: Optional[np.ndarray] = None,
    snr: Optional[np.ndarray] = None,
    snr_mask: Optional[np.ndarray] = None,
    cells: Optional[np.ndarray] = None,
    output: str = "iscell_curated.npy",
):

    with napari.gui_qt():
        viewer = napari.view_image(
            img, multiscale=False, contrast_limits=data_range, visible=True
        )

        # add the SNR widgets
        if (snr_mask is not None) and (snr is not None):
            snr_image = viewer.add_image(snr_mask, visible=False)
            snr_clean = np.nan_to_num(snr, copy=True, neginf=0, nan=0)
            bins = np.linspace(
                np.rint(snr_clean.min()), np.rint(snr_clean.max()), 20
            )
            y, x = np.histogram(snr, bins=bins)
            hist = HistogramWidget(x, y, xlabel="SNR", ylabel="counts")

            def snr_dragged(event=None):
                snr_thresh = hist.hist_plot._vert_line.getPos()[0]
                if np.sum(contour_manager.good_contour) > 0:
                    good_snr_indices = np.squeeze(
                        np.argwhere(
                            snr[contour_manager.good_contour] > snr_thresh
                        )
                    )
                    if good_snr_indices.ndim != 0:
                        good_masks = [
                            contour_manager.accepted_contours[i]
                            for i in good_snr_indices
                        ]
                        good_snr = snr[contour_manager.good_contour][
                            good_snr_indices
                        ]
                        im_shape = (img.shape[-2], img.shape[-1])
                        new_snr_mask = make_scalar_mask(
                            good_masks, im_shape, good_snr
                        )
                        snr_image.data = new_snr_mask

            # update the SNR image and connect the event
            snr_dragged()
            hist.hist_plot.connect_line_dragged(snr_dragged)

            viewer.window.add_dock_widget(hist, name="SNR histogram")

        # Add the cell labels
        selected_shapes = viewer.add_shapes(name="selected_cell")
        rejected_mask = contour_manager.make_rejected_mask()
        bad_labels = viewer.add_labels(rejected_mask, visible=False)

        accepted_mask = contour_manager.make_accepted_mask()
        good_labels = viewer.add_labels(accepted_mask)

        t = np.arange(len(f[0]))
        if spikes is not None:
            spike_events = spikes[0] > 50
        else:
            spike_events = None
        line_plot = LinePlot(
            x=t,
            y=f[0],
            xlabel="time",
            ylabel="fluorescence",
            events=spike_events,
        )
        viewer.window.add_dock_widget(line_plot, name="Fluorescence trace")

        def update_line(event=None):
            current_frame = viewer.dims.point[0]
            line_plot.update_vline(current_frame)

        viewer.dims.events.axis.connect(update_line)

        def update_plot(cell_indices):
            for i in cell_indices:
                data = f[i]
                if spikes is not None:
                    spike_events = spikes[i] > 25
                else:
                    spike_events = None
                line_plot.plot(data, spike_events)
                current_frame = viewer.dims.point[0]
                line_plot.update_vline(current_frame)

        def update_selection(selected_index):
            if selected_index is not None:
                selected_contour = selected_index - 1

                # clear any current selections
                contour_manager.selected_contours = {}
                selected_shapes.selected_data = np.arange(
                    len(selected_shapes.data)
                )
                selected_shapes.remove_selected()

                if selected_contour != -1:
                    contour_manager.selected_contours = {selected_contour}
                    selection_bbox = []
                    contour = contour_manager.contours[selected_contour]
                    min_r = np.min(contour[:, 0])
                    min_c = np.min(contour[:, 1])
                    max_r = np.max(contour[:, 0])
                    max_c = np.max(contour[:, 1])

                    selection_bbox.append(
                        np.array(
                            [
                                [min_r, min_c],
                                [min_r, max_c],
                                [max_r, max_c],
                                [max_r, min_c],
                            ]
                        )
                    )
                    selected_shapes.add(
                        selection_bbox,
                        shape_type="rectangle",
                        face_color="transparent",
                        edge_color="green",
                    )
                    update_plot(cell_indices=[selected_contour])

            else:
                line_plot.clear()

        def select_on_click(viewer, event):
            selected_layers = viewer.layers.selected
            if len(selected_layers) == 1:

                if isinstance(selected_layers[0], napari.layers.Labels):
                    selected_index = selected_layers[0]._value
                    update_selection(selected_index)
                elif (snr_mask is not None) and (snr is not None):
                    if selected_layers[0] is snr_image:
                        visual = viewer.window.qt_viewer.layer_to_visual[
                            good_labels
                        ]
                        pos = list(event.pos)
                        good_labels.position = visual._transform_position(pos)
                        selected_index = good_labels._get_value()

                        if selected_index is not None:
                            update_selection(selected_index)
            yield

        viewer.mouse_drag_callbacks.append(select_on_click)

        @viewer.bind_key("t")
        def toggle_selected_contour(viewer):
            selected_contours = list(contour_manager.selected_contours)
            if len(selected_contours) > 0:
                good_contour = contour_manager.good_contour
                good_contour[selected_contours] = ~good_contour[
                    selected_contours
                ]
                contour_manager.good_contour = good_contour

                good_contours_image = contour_manager.make_accepted_mask()
                bad_contours_image = contour_manager.make_rejected_mask()

                good_labels.data = good_contours_image
                bad_labels.data = bad_contours_image

                if (snr_mask is not None) and (snr is not None):
                    snr_dragged()

        @viewer.bind_key("q")
        def save_cells(viewer):
            snr_thresh = hist.hist_plot._vert_line.getPos()[0]
            accepted_cells_indices = np.argwhere(
                snr[contour_manager.good_contour] > snr_thresh
            )
            accepted_cells = np.zeros((f.shape[0],))
            accepted_cells[accepted_cells_indices] = 1
            if cells is not None:
                good_cells = np.hstack(
                    (
                        np.expand_dims(accepted_cells, 1),
                        np.expand_dims(cells[:, 1], 1),
                    )
                )
            else:
                good_cells = accepted_cells
            np.save(output, good_cells)

        viewer.show()



from typing import Tuple

import numpy as np


def make_scalar_mask(
    masks, im_shape: Tuple[int, int], values: np.ndarray
) -> np.ndarray:
    mask_im = np.zeros(im_shape)

    for mask, value in zip(masks, values):
        mask_im[
            np.round(mask[:, 0]).astype("int"),
            np.round(mask[:, 1]).astype("int"),
        ] = value

    return mask_im



import os
import pathlib
from typing import Any, Dict, Tuple

import h5py
import numpy as np
from scipy import sparse

"""
Code from Caiman utils

https://github.com/flatironinstitute/CaImAn/blob/e2de6f3350342f2f5762279b7059bdb739d58d84/caiman/utils/utils.py


"""

# \package Caiman/utils
# \version   1.0
# \bug
# \warning
# \copyright GNU General Public License v2.0
# \date Created on Tue Jun 30 21:01:17 2015
# \author: andrea giovannucci
# \namespace utils
# \pre none


def load_dict_from_hdf5(filename: str) -> Dict:
    """ Load dictionary from hdf5 file
    Args:
        filename: str
            input file to load
    Returns:
        dictionary
    """

    with h5py.File(filename, "r") as h5file:
        return recursively_load_dict_contents_from_group(h5file, "/")


def recursively_load_dict_contents_from_group(
    h5file: h5py.File, path: str
) -> Dict:
    """load dictionary from hdf5 object
    Args:
        h5file: hdf5 object
            object where dictionary is stored
        path: str
            path within the hdf5 file
    """

    ans: Dict = {}
    for key, item in h5file[path].items():

        if isinstance(item, h5py._hl.dataset.Dataset):
            val_set = np.nan
            if isinstance(item[()], str):
                if item[()] == "NoneType":
                    ans[key] = None
                else:
                    ans[key] = item[()]

            elif key in [
                "dims",
                "medw",
                "sigma_smooth_snmf",
                "dxy",
                "max_shifts",
                "strides",
                "overlaps",
            ]:

                if type(item[()]) == np.ndarray:
                    ans[key] = tuple(item[()])
                else:
                    ans[key] = item[()]
            else:
                if type(item[()]) == np.bool_:
                    ans[key] = bool(item[()])
                else:
                    ans[key] = item[()]

        elif isinstance(item, h5py._hl.group.Group):
            if key in ("A", "W", "Ab", "downscale_matrix", "upscale_matrix"):
                data = item[path + key + "/data"]
                indices = item[path + key + "/indices"]
                indptr = item[path + key + "/indptr"]
                shape = item[path + key + "/shape"]
                ans[key] = sparse.csc_matrix(
                    (data[:], indices[:], indptr[:]), shape[:]
                )
                if key in ("W", "upscale_matrix"):
                    ans[key] = ans[key].tocsr()
            else:
                ans[key] = recursively_load_dict_contents_from_group(
                    h5file, path + key + "/"
                )
    return ans


def prepare_shape(mytuple: Tuple) -> Tuple:
    """ This promotes the elements inside a shape into np.uint64. It is intended to prevent overflows
        with some numpy operations that are sensitive to it, e.g. np.memmap """
    if not isinstance(mytuple, tuple):
        raise Exception("Internal error: prepare_shape() passed a non-tuple")
    return tuple(map(lambda x: np.uint64(x), mytuple))


def load_memmap(filename: str, mode: str = "r") -> Tuple[Any, Tuple, int]:
    """ Load a memory mapped file created by the function save_memmap
    Args:
        filename: str
            path of the file to be loaded
        mode: str
            One of 'r', 'r+', 'w+'. How to interact with files
    Returns:
        Yr:
            memory mapped variable
        dims: tuple
            frame dimensions
        T: int
            number of frames
    Raises:
        ValueError "Unknown file extension"
    """
    if pathlib.Path(filename).suffix != ".mmap":
        raise ValueError("Unknown file extension (should be .mmap)")
    # Strip path components and use CAIMAN_DATA/example_movies
    # TODO: Eventually get the code to save these in a different dir
    file_to_load = filename
    filename = os.path.split(filename)[-1]
    fpart = filename.split("_")[
        1:-1
    ]  # The filename encodes the structure of the map
    d1, d2, d3, T, order = (
        int(fpart[-9]),
        int(fpart[-7]),
        int(fpart[-5]),
        int(fpart[-1]),
        fpart[-3],
    )
    Yr = np.memmap(
        file_to_load,
        mode=mode,
        shape=prepare_shape((d1 * d2 * d3, T)),
        dtype=np.float32,
        order=order,
    )
    if d3 == 1:
        return (Yr, (d1, d2), T)
    else:
        return (Yr, (d1, d2, d3), T)



import argparse

from .calcium_curator import calcium_curator
from .io.caiman.caiman_reader import caiman_reader
from .io.s2p.s2p_reader import s2p_reader


READER_FUNCS = {"s2p": s2p_reader, "caiman": caiman_reader}


def parse_args():
    parser = argparse.ArgumentParser(description="CalciumCurator")
    parser.add_argument("--pipeline", default="", type=str, help="options")
    parser.add_argument(
        "--pipeline-params", default="", type=str, help="options"
    )
    parser.add_argument("--image", default="", type=str, help="options")
    parser.add_argument("--snr", default="", type=str, help="options")
    parser.add_argument("--trace", default="", type=str, help="options")
    parser.add_argument("--cell", default="", type=str, help="options")
    parser.add_argument("--spikes", default="", type=str, help="options")
    parser.add_argument("--output", default=".", type=str, help="options")

    args = parser.parse_args()

    pipeline_name = args.pipeline
    pipeline_params = args.pipeline_params
    image_path = args.image
    snr_path = args.snr
    trace_path = args.trace
    cell_path = args.cell
    spikes_path = args.spikes
    output_dir = args.output

    return (
        pipeline_name,
        pipeline_params,
        image_path,
        snr_path,
        trace_path,
        cell_path,
        spikes_path,
        output_dir,
    )


def main():
    (
        pipeline_name,
        pipeline_params,
        image_path,
        snr_path,
        trace_path,
        cell_path,
        spikes_path,
        output_dir,
    ) = parse_args()

    try:
        reader_func = READER_FUNCS[pipeline_name]
    except KeyError:
        raise KeyError("unknown pipline. valid options are s2p and caiman")

    (
        im_registered,
        data_range,
        contour_manager,
        f_traces,
        snr,
        snr_mask,
        spikes,
        is_cell,
    ) = reader_func(
        pipeline_params,
        image_path,
        snr_path,
        trace_path,
        cell_path,
        spikes_path,
    )

    calcium_curator(
        img=im_registered,
        data_range=data_range,
        contour_manager=contour_manager,
        f=f_traces,
        snr=snr,
        snr_mask=snr_mask,
        spikes=spikes,
        output=output_dir,
        cells=is_cell,
    )



import argparse
import os

from .calcium_curator import calcium_curator
from .io.caiman.caiman_reader import caiman_reader
from .io.caiman._vendored import load_dict_from_hdf5


def parse_args():
    parser = argparse.ArgumentParser(description="view-caiman")
    parser.add_argument("--results", default="", type=str, help="options")
    parser.add_argument("--image", default="", type=str, help="options")
    parser.add_argument("--output", default="", type=str, help="options")

    args = parser.parse_args()
    results_file = args.results
    image_path = args.image
    output_dir = args.output

    return results_file, image_path, output_dir


def view_caiman():
    results_file, image_path, output_dir = parse_args()

    cnm_obj = load_dict_from_hdf5(results_file)

    if image_path == "":
        im_name = os.path.basename(cnm_obj["mmap_file"])
        results_dir = os.path.dirname(results_file)
        image_path = os.path.join(results_dir, im_name)
        if not os.path.isfile(image_path):
            raise FileNotFoundError(
                "Image file not found in results directory.\n"
                "Try passing the image file path with the --image argument"
            )

    (
        im_registered,
        data_range,
        contour_manager,
        f_traces,
        snr,
        snr_mask,
        spikes,
        is_cell,
    ) = caiman_reader(results_file, image_path)

    calcium_curator(
        img=im_registered,
        data_range=data_range,
        contour_manager=contour_manager,
        f=f_traces,
        snr=snr,
        snr_mask=snr_mask,
        spikes=spikes,
        output=output_dir,
        cells=is_cell,
    )



from typing import Union

import numpy as np
from ..utils.data_range import calc_data_range
from skimage import measure

from ...contour_manager import ContourManager
from ...images.masks import make_scalar_mask
from ._vendored import load_dict_from_hdf5, load_memmap


def make_caiman_contour_manager(
    img_components: np.ndarray, good_indices: Union[list, np.ndarray]
) -> ContourManager:
    contours = [measure.find_contours(comp, 40)[0] for comp in img_components]
    initial_state = np.zeros((len(contours),), dtype=np.bool)
    initial_state[good_indices] = True
    contour_manager = ContourManager(
        contours,
        initial_state=initial_state,
        im_shape=img_components[0, ...].shape,
    )

    return contour_manager


def load_movie(filename: str):
    """Adapted from caiman


    """
    # filename = os.path.basename(filename)
    Yr, dims, T = load_memmap(filename)
    images = np.reshape(Yr.T, [T] + list(dims), order="C")

    return images


def caiman_reader(
    pipeline_params,
    image_path,
    snr_path=None,
    trace_path=None,
    cell_path=None,
    spikes_path=None,
):

    # Load the image
    im_registered = load_movie(image_path)
    data_range = calc_data_range(im_registered)

    # load the pipeline output object
    cnm_obj = load_dict_from_hdf5(pipeline_params)

    # make the contours
    estimates = cnm_obj["estimates"]
    if estimates["dims"] is None:
        estimates["dims"] = im_registered.shape[1::]
    img_components = (
        estimates["A"]
        .toarray()
        .reshape((estimates["dims"][0], estimates["dims"][1], -1), order="F")
        .transpose([2, 0, 1])
    )
    img_components = (
        img_components / img_components.max(axis=(1, 2))[:, None, None]
    )
    img_components = img_components * 255
    estimates["img_components"] = img_components.astype(np.uint8)
    contour_manager = make_caiman_contour_manager(
        estimates["img_components"], good_indices=estimates["idx_components"]
    )

    # calculate the SNR and make the mask
    snr = estimates["SNR_comp"]
    im_shape = im_registered.shape
    contours = [
        measure.find_contours(comp, 40)[0]
        for comp in estimates["img_components"]
    ]
    snr_mask = make_scalar_mask(
        contours, im_shape=(im_shape[-2], im_shape[-1]), values=snr
    )

    # get the fluorescence data
    f_traces = estimates["C"] + estimates["YrA"]

    # caiman doesn't use spikes and is_cell for now
    is_cell = None
    spikes = None

    return (
        im_registered,
        data_range,
        contour_manager,
        f_traces,
        snr,
        snr_mask,
        spikes,
        is_cell,
    )
